class TinygraphioError(Exception):
    code: int

class UnknownError(TinygraphioError):
    code = 1

class InternalError(TinygraphioError):
    code = 2

class DataLossError(TinygraphioError):
    code = 3

class InvalidArgumentError(TinygraphioError):
    code = 4



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tinygraphio.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11tinygraphio.proto\x12\x0etinygraphio.v1\"/\n\nFileHeader\x12\x0f\n\x07version\x18\x01 \x02(\x06\x12\x10\n\x08\x63hecksum\x18\x02 \x02(\x06\"+\n\x0bGraphHeader\x12\r\n\x05nodes\x18\x01 \x02(\x06\x12\r\n\x05\x65\x64ges\x18\x02 \x02(\x06\"6\n\nGraphChunk\x12\x13\n\x07offsets\x18\x01 \x03(\x04\x42\x02\x10\x01\x12\x13\n\x07targets\x18\x02 \x03(\x04\x42\x02\x10\x01\x42\x02H\x03')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tinygraphio_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'H\003'
  _GRAPHCHUNK.fields_by_name['offsets']._options = None
  _GRAPHCHUNK.fields_by_name['offsets']._serialized_options = b'\020\001'
  _GRAPHCHUNK.fields_by_name['targets']._options = None
  _GRAPHCHUNK.fields_by_name['targets']._serialized_options = b'\020\001'
  _FILEHEADER._serialized_start=37
  _FILEHEADER._serialized_end=84
  _GRAPHHEADER._serialized_start=86
  _GRAPHHEADER._serialized_end=129
  _GRAPHCHUNK._serialized_start=131
  _GRAPHCHUNK._serialized_end=185
# @@protoc_insertion_point(module_scope)



from tinygraphio.reader import TinygraphioReader


def main(args):
    with args.tinygraph.open("rb") as f:
        reader = TinygraphioReader(f)
        header = reader._read_file_header()

        print(header.version)



from typing import Optional

from google.protobuf.internal.decoder import _DecodeVarint

from tinygraphio.graph import Tinygraph, Node, Edge
from tinygraphio.errors import DataLossError
from tinygraphio.utils import delta_decode

from tinygraphio.proto.tinygraphio_pb2 import FileHeader, GraphHeader, GraphChunk


class TinygraphioReader:
    def __init__(self, io):
        self.io = io

    def read(self) -> Tinygraph:
        return self._read_graph()

    def close(self) -> None:
        self.io.close()

    def _read_file_header(self) -> FileHeader:
        # 2x fixed64 plus their tag
        buf = self.io.read(2 * 8 + 2)

        if not buf:
            raise DataLossError("Could not read all file header bytes")

        obj = FileHeader()
        obj.ParseFromString(buf)

        return obj

    def _read_graph_header(self) -> GraphHeader:
        # 2x fixed64 plus their tag
        buf = self.io.read(2 * 8 + 2)

        if not buf:
            raise DataLossError("Could not read all graph header bytes")

        obj = GraphHeader()
        obj.ParseFromString(buf)

        return obj

    def _read_graph_chunk(self) -> Optional[GraphChunk]:
        # TODO: this will fail if we write out e.g. an
        # empty graph chunk and a prefix size of 0. We
        # should in addition store number of graph chunks
        # e.g. in the graph header or use a fixed size
        # prefix here instead of a variable sized varint.
        szbuf = self.io.peek(4)

        if not szbuf:
            return None

        size, pos = _DecodeVarint(szbuf, pos=0)
        self.io.seek(pos, 1)

        buf = self.io.read(size)

        if not buf:
            raise DataLossError("Could not read all graph chunk bytes")

        obj = GraphChunk()
        obj.ParseFromString(buf)

        return obj

    def _read_graph(self) -> Tinygraph:
        file_header = self._read_file_header()
        graph_header = self._read_graph_header()

        offsets = []
        targets = []

        while True:
            chunk = self._read_graph_chunk()

            if not chunk:
                break

            offsets += chunk.offsets
            targets += chunk.targets

        return Tinygraph(offsets=[Edge(e) for e in delta_decode(offsets)],
                         targets=[Node(n) for n in targets])

    def __repr__(self) -> str:
        return f"<TinygraphioReader {self.io}>"



import argparse
from pathlib import Path

import tinygraphio.scripts.version


def main():
    parser = argparse.ArgumentParser(prog="tinygraphio")
    Formatter = argparse.ArgumentDefaultsHelpFormatter

    subcmd = parser.add_subparsers(dest="command")
    subcmd.required = True

    version = subcmd.add_parser("version", formatter_class=Formatter)
    version.add_argument("tinygraph", type=Path, help="path to tinygraph")
    version.set_defaults(main=tinygraphio.scripts.version.main)

    args = parser.parse_args()
    args.main(args)


if __name__ == "__main__":
    main()



from typing import NamedTuple, NewType
from collections.abc import Iterable, Sequence


Node = NewType("Node", int)
Edge = NewType("Edge", int)

class Tinygraph(NamedTuple):
    offsets: Sequence[Edge]
    targets: Sequence[Node]

    def num_nodes(self) -> int:
        return max(0, len(self.offsets) - 1)

    def num_edges(self) -> int:
        return len(self.targets)

    def is_empty(self) -> bool:
        return self.num_nodes() == 0;

    def edges(self, n: Node) -> Iterable[Edge]:
        return [Edge(e) for e in range(self.offsets[n], self.offsets[n + 1])]

    def target(self, e: Edge) -> Node:
        return self.targets[e]

    def degree(self, n: Node) -> int:
        return self.offsets[n + 1] - self.offsets[n]

    def neighbors(self, n: Node) -> Iterable[Node]:
        return [self.target(e) for e in self.edges(n)]

    def has_node(self, n: Node) -> bool:
        return 0 <= n < self.num_nodes()

    def has_edge(self, e: Edge) -> bool:
        return 0 <= e < self.num_edges()

    def has_edge_from_to(self, s: Node, t: Node) -> bool:
        return t in set(self.neighbors(s))

    def __repr__(self) -> str:
        return f"<Tinygraph nodes={self.num_nodes()}, edges={self.num_edges()}>"



from google.protobuf.internal.encoder import _VarintBytes, _VarintSize

from tinygraphio.graph import Tinygraph
from tinygraphio.errors import DataLossError
from tinygraphio.utils import delta_encode

from tinygraphio.proto.tinygraphio_pb2 import FileHeader, GraphHeader, GraphChunk


class TinygraphioWriter:
    def __init__(self, io):
        self.io = io

    def write(self, graph: Tinygraph) -> int:
        return self._write_graph(graph)

    def flush(self) -> None:
        self.io.flush()

    def close(self) -> None:
        self.io.close()

    def _write_file_header(self, file_header: FileHeader) -> int:
        assert file_header.IsInitialized()

        buf = file_header.SerializeToString()
        assert len(buf) == (2 * 8 + 2)

        if self.io.write(buf) != len(buf):
            raise DataLossError("Could not write all file header bytes")

        return len(buf)

    def _write_graph_header(self, graph_header: GraphHeader) -> int:
        assert graph_header.IsInitialized()

        buf = graph_header.SerializeToString()
        assert len(buf) == (2 * 8 + 2)

        if self.io.write(buf) != len(buf):
            raise DataLossError("Could not write all graph header bytes")

        return len(buf)

    def _write_graph_chunk(self, graph_chunk: GraphChunk) -> int:
        assert graph_chunk.IsInitialized()

        buf = graph_chunk.SerializeToString()

        n = self.io.write(_VarintBytes(len(buf)) + buf)

        if n != _VarintSize(len(buf)) + len(buf):
            raise DataLossError("Could not write all graph chunk bytes")

        return n

    def _write_graph(self, graph: Tinygraph) -> int:
        file_header = FileHeader(version=1, checksum=0)

        graph_header = GraphHeader(nodes=graph.num_nodes(),
                                   edges=graph.num_edges())

        written = 0
        written += self._write_file_header(file_header)
        written += self._write_graph_header(graph_header)

        # The chunk size is a trade-off between having to decode
        # one chunk in memory at a time vs the overhead of chunks.
        # With a size ~20k we will get chunks of ~40k-160k bytes,
        # depending on how small varints will be encoded.
        chunk_size = 20000

        n = max(len(graph.offsets), len(graph.targets))

        for i in range(0, n, chunk_size):
            offsets = delta_encode(graph.offsets[i : i + n])
            targets = graph.targets[i : i + n]

            graph_chunk = GraphChunk(offsets=offsets, targets=targets)

            written += self._write_graph_chunk(graph_chunk)

        return written

    def __repr__(self) -> str:
        return f"<TinygraphioWriter {self.io}>"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tinygraphio', 'tinygraphio.proto', 'tinygraphio.scripts']

package_data = \
{'': ['*']}

install_requires = \
['protobuf>=4.21.12,<5.0.0']

entry_points = \
{'console_scripts': ['tinygraphio = tinygraphio.scripts.__main__:main']}

setup_kwargs = {
    'name': 'tinygraphio',
    'version': '1.0.0',
    'description': 'The tinygraphio graph data interchange file format',
    'long_description': '# tinygraphio\n\nPython implementation of the tinygraphio graph data interchange file format.\n\n\n## Installation\n\nInstall with `poetry`\n\n    poetry add tinygraphio\n\nInstall with `pip`\n\n    pip install tinygraphio\n\n\n## Usage\n\n- The `Tinygraph` class implements a compressed sparse row graph\n- The `TinygraphioReader` implements reading a graph from a binary file-like object\n- The `TinygraphioWriter` implements writing a graph to a binary file-like object\n\n```python3\nfrom tinygraphio.graph import Tinygraph, Node, Edge\nfrom tinygraphio.reader import TinygraphioReader\nfrom tinygraphio.writer import TinygraphioWriter\n```\n\nWriting\n\n```\ngraph = Tinygraph(offsets=[0, 2, 4, 5], targets=[1, 2, 0, 2, 1])\n\nwith open("berlin.tinygraph", "wb") as f:\n    writer = TinygraphioWriter(f)\n    writer.write(graph)\n```\n\nReading\n\n```\nwith open("berlin.tinygraph", "rb") as f:\n    reader = TinygraphioReader(f)\n    graph = reader.read()\n```\n\nNote: this library implements reading and writing a compressed sparse row graph in a \nThe use case tinygraphio covers is storing large graphs effectively and efficiently and sharing graphs in a portable way.\nWe do not provide a full-blown graph computation toolkit on purpose.\n\n## Development\n\n\n## License\n\nCopyright © 2023 tinygraph\n\nDistributed under the MIT License (MIT).\n',
    'author': 'tinygraph',
    'author_email': 'hello@tinygraph.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>3.8,<=4',
}


setup(**setup_kwargs)



from collections.abc import Iterable


def delta_encode(it: Iterable[int], prev: int = 0) -> Iterable[int]:
    r = []

    for v in it:
        r.append(v - prev)
        prev = v

    return r

def delta_decode(it: Iterable[int], prev: int = 0) -> Iterable[int]:
    r = []

    for v in it:
        r.append(v + prev)
        prev += v

    return r
